# Copyright 2025 Sear Gamemode
import mne
import numpy as np
from scipy.spatial import KDTree
from scipy.spatial import Delaunay

def align_head(ch_dict, nasion, lpa, rpa, hsp, standard='waveguard64', mode='Cz', threshold=0.1):
    def calculate_rotation_matrix(A, B):
        H = A @ B.T
        U, S, Vt = np.linalg.svd(H)
        R = Vt.T @ U.T
        return R
    
    def interpolate_electrode_proj(pos, neighbor, hsp):
        tree = KDTree(hsp)
        _, idx = tree.query(neighbor, k=1)
        return hsp[idx]
    
    def interpolate_electrode_tri(ch_dict, standard_pos, hsp):
        tri = Delaunay(hsp)
        simplex = tri.find_simplex(standard_pos)
        if simplex == -1:
            print(f"Electrode {standard_pos} is outside the triangulation")
        interpolated_pos = np.mean(hsp[simplex >= 0], axis=0)
        return interpolated_pos

    def coord_update(ch_dict, nasion, lpa, rpa, hsp, vector):
        ch_dict = {ch: pos+vector for ch, pos in ch_dict.items()} 
        nasion += vector
        lpa += vector
        rpa += vector
        hsp += vector
        return ch_dict, nasion, lpa, rpa, hsp
    def compute_stats(ch_dict):
        coords = np.array(list(ch_dict.values()))
        mean = np.mean(coords, axis=0)
        std = np.std(coords, axis=0)
        return mean, std
    standard_montage = create_custom_montage(standard)
    standard_pos = standard_montage.get_positions()['ch_pos']
    standard_points = np.array([standard_pos['Cz'], standard_pos['M1'], standard_pos['M2'], standard_pos['Fpz']])
    points = np.array([ch_dict['Cz'], ch_dict['M1'], ch_dict['M2'], ch_dict['Fpz']])
    R = calculate_rotation_matrix(points.T, standard_points.T)

    ch_dict = {ch: R.dot(pos) for ch, pos in ch_dict.items()}
    translation_vector = np.mean(list(standard_pos.values())) - np.mean(list(ch_dict.values()))
    nasion = R.dot(nasion)
    lpa = R.dot(lpa)
    rpa = R.dot(rpa)
    hsp = np.dot(hsp, R.T)
    ch_dict, nasion, lpa, rpa, hsp, = coord_update(ch_dict, nasion, lpa, rpa, hsp, translation_vector)

    ind_mean, ind_std = compute_stats(ch_dict)
    st_mean, st_std = compute_stats(standard_pos)

    ch_norm = {ch: ((pos-ind_mean)/ind_std)*st_std + st_mean for ch, pos in ch_dict.items()}
    hsp_norm = np.array([((coord-ind_mean)/ind_std)*st_std + st_mean for coord in hsp])

    for ch in standard_pos.keys():
        ch_error = np.linalg.norm(ch_norm[ch] - standard_pos[ch])
        #print(ch, ch_error)
        if  ch_error > threshold:
            print(f'{ch} interpolated with error={ch_error}')
            #neighbors_individual = np.array([ch_dict[n_ch] for n_ch in ch_dict if n_ch != ch])
            ch_coord = interpolate_electrode_proj(ch_norm[ch], standard_pos[ch], hsp_norm)
            #ch_coord = interpolate_electrode_tri(ch_norm, standard_pos[ch], hsp_norm)
            ch_error = np.linalg.norm(ch_coord- standard_pos[ch])
            print(f'New error={ch_error}')
            ch_dict[ch] = ((ch_coord-st_mean)/st_std)*ind_std + ind_mean
    
    if mode == 'Cz':
        translation_vector = standard_pos['Cz'] - ch_dict['Cz']
        ch_dict, nasion, lpa, rpa, hsp, = coord_update(ch_dict, nasion, lpa, rpa, hsp, translation_vector)
    return ch_dict, nasion, lpa, rpa, hsp

def read_elc(elc_file):
    lines = []
    with open(elc_file, 'r', encoding='utf-8') as file:
        for line in file:
            lines.append(line.strip())

    num_ch = int(lines[0].strip().split('=')[1])
    ch_dict = {}
    for line in lines[3:num_ch+3]:
        ch, coord = line.split(':')
        ch = ch.replace('\t', '')
        coord = list(map(float, coord.split('\t')[1:]))
        coord = [-coord[1]/1000, coord[0]/1000, coord[2]/1000]
        ch_dict[ch]=coord

    num_hsp = int(lines[num_ch+5].strip().split('=')[1])
    hsp = []
    for line in lines[num_ch+8:num_ch+8+num_hsp]:
        coord = list(map(float, line.split('\t')))
        coord = [-coord[1]/1000, coord[0]/1000, coord[2]/1000]
        hsp.append(coord)

    nasion = ch_dict['Nasion']
    lpa = ch_dict['LeftEar']
    rpa = ch_dict['RightEar']
    del ch_dict['EOG']
    del ch_dict['Nasion']
    del ch_dict['LeftEar']
    del ch_dict['RightEar']
    return ch_dict, nasion, lpa, rpa, hsp

def create_custom_montage(montage):
    if montage == 'waveguard64':
        ch_pos = {
        'Fp1': np.array([94.9, 30.7, 14.0]),
        'Fpz': np.array([98.3, -0.2, 18.6]),
        'Fp2': np.array([96.4, -31.5, 14.1]),
        'F7' : np.array([52.0, 74.2, 3.1]),
        'F3' : np.array([52.0, 52.9, 56.7]),
        'Fz' : np.array([52.4, -0.4, 79.2]),
        'F4' : np.array([53.1, -54.9, 55.4]),
        'F8' : np.array([53.6, -76.0, 2.9]),
        'FC5': np.array([18.5, 80.9, 33.5]),
        'FC1': np.array([15.1, 34.7, 87.3]),
        'FC2': np.array([15.8, -35.8, 85.8]),
        'FC6': np.array([19.9, -84.0, 33.3]),
        'M1' : np.array([-37.1, 89.8, -68.3]),
        'T7' : np.array([-11.7, 87.4, -6.2]),
        'C3' : np.array([-22.8, 67.2, 64.9]),
        'Cz' : np.array([-27.5, -0.5, 96.9]),
        'C4' : np.array([-21.9, -69.1, 64.0]),
        'T8' : np.array([-10.6, -88.2, -6.3]),
        'M2' : np.array([-36.9, -88.8, -69.2]),
        'CP5': np.array([-53.1, 82.8, 26.0]),
        'CP1': np.array([-66.8, 37.1, 83.7]),
        'CP2': np.array([-66.3, -39.9, 83.6]),
        'CP6': np.array([-52.4, -86.2, 26.7]),
        'P7' : np.array([-74.9, 75.6, -12.0]),
        'P3' : np.array([-92.6, 55.1, 42.5]),
        'Pz' : np.array([-100.5, -0.4, 66.9]),
        'P4' : np.array([-92.7, -57.3, 42.6]),
        'P8' : np.array([-74.9, -76.2, -12.2]),
        'POz': np.array([-115.9, -0.3, 30.7]),
        'O1' : np.array([-118.8, 30.8, -11.0]),
        'Oz' : np.array([-122.6, -0.2, -7.3]),
        'O2' : np.array([-118.8, -31.2, -11.1]),
        'AF7': np.array([79.3, 57.4, 8.6]),
        'AF3': np.array([82.4, 35.5, 41.0]),
        'AF4': np.array([83.1, -37.7, 41.4]),
        'AF8': np.array([80.3, -57.9, 8.5]),
        'F5' : np.array([52.0, 67.4, 31.3]),
        'F1' : np.array([52.1, 28.5, 74.1]),
        'F2' : np.array([52.7, -31.0, 73.2]),
        'F6' : np.array([53.4, -70.0, 30.8]),
        'FC3': np.array([16.5, 63.1, 64.4]),
        'FCz': np.array([14.8, -0.5, 95.5]),
        'FC4': np.array([17.6, -64.7, 63.4]),
        'C5' : np.array([-17.6, 84.9, 31.6]),
        'C1' : np.array([-26.7, 38.0, 90.8]),
        'C2' : np.array([-26.0, -39.8, 89.2]),
        'C6' : np.array([-16.6, -86.3, 31.4]),
        'CP3': np.array([-61.0, 65.7, 58.9]),
        'CPz': np.array([-68.2, -0.5, 90.2]),
        'CP4': np.array([-60.1, -68.0, 58.5]),
        'P5' : np.array([-84.3, 70.5, 17.0]),
        'P1' : np.array([-98.3, 29.6, 60.2]),
        'P2' : np.array([-98.6, -32.1, 60.8]),
        'P6' : np.array([-84.3, -72.2, 16.9]),
        'PO5': np.array([-107.2, 50.3, 4.9]),
        'PO3': np.array([-112.3, 39.6, 20.4]),
        'PO4': np.array([-112.5, -38.3, 19.4]),
        'PO6': np.array([-107.5, -49.5, 4.1]),
        'FT7': np.array([20.9, 83.6, -1.8]),
        'FT8': np.array([22.4, -85.0, -2.0]),
        'TP7': np.array([-44.5, 87.9, -10.1]),
        'TP8': np.array([-43.5, -87.0, -10.0]),
        'PO7': np.array([-101.5, 57.2, -12.6]),
        'PO8': np.array([-102.1, -58.2, -12.9]),
        }
    for key in ch_pos.keys():
        ch_pos[key][0], ch_pos[key][1] = -ch_pos[key][1], ch_pos[key][0]
        ch_pos[key] /= 1000
    Nasion= [ 5.27205792e-18,  8.60992398e-02, -4.01487349e-02]
    LPA= [-0.08609924, -0., -0.04014873]
    RPA= [ 0.08609924,  0., -0.04014873]
    dig_montage = mne.channels.make_dig_montage(ch_pos=ch_pos, nasion=Nasion, lpa=LPA, rpa=RPA, coord_frame='head')
    return dig_montage



# Copyright 2025 Sear Gamemode
import seaborn as sns
import matplotlib.pyplot as plt
import mne 
import pandas as pd 
import os 
import numpy as np 
from matplotlib.gridspec import GridSpec
from matplotlib.colors import ListedColormap
from matplotlib.lines import Line2D

def plot_clusters(clusters, clusters_avg, clusters_max, clusters_min, ch_names, ch_pos, correlation_values):
    figs = []
    # Create an array of data where each channel is assigned a cluster value
    data = np.full(len(ch_names), np.nan)
    # Create a mask for electrodes in clusters
    mask = np.zeros(len(ch_names), dtype=bool)
    # Map channel names to indices
    ch_name_to_idx = {ch: idx for idx, ch in enumerate(ch_names)}

    for cluster_idx, cluster in enumerate(clusters):
        cluster_value = cluster_idx + 1  # Cluster values start from 1
        for ch_name in cluster:
            if ch_name in ch_name_to_idx:
                idx = ch_name_to_idx[ch_name]
                data[idx] = cluster_value
                mask[idx] = True  # Mark the electrode as belonging to a cluster

    # Create a custom colormap
    N_clusters = len(clusters)
    cmap = plt.get_cmap('tab20', N_clusters)
    cluster_colors = cmap(np.arange(N_clusters) % 20)
    cluster_cmap = ListedColormap(cluster_colors)

    # Display the topomap
    fig, ax = plt.subplots(figsize=(8, 8))
    im, cn = mne.viz.plot_topomap(
        data,
        pos=ch_pos,
        axes=ax,
        cmap=cluster_cmap,
        names=ch_names,
        show=False,
        vlim=(0.5, N_clusters + 0.5),
        contours=0,
        image_interp='nearest',
        extrapolate='head',
        outlines='head',
        sphere=(0., 0., 0., 0.095)
    )

    ax.set_title('Bridged Clusters')

    # Create a custom legend
    legend_elements = []
    for idx, (avg_corr, max_corr, min_corr) in enumerate(zip(clusters_avg, clusters_max, clusters_min)):
        color = cluster_colors[idx % 20]
        label = f'Cluster {idx+1}: Correlation (avg={avg_corr:.3f}, max={max_corr:.3f}, min={min_corr:.3f})'
        legend_elements.append(Line2D(
            [0], [0],
            marker='s',
            color='w',
            label=label,
            markerfacecolor=color,
            markersize=10
        ))

    ax.legend(
        handles=legend_elements,
        bbox_to_anchor=(1.05, 1),
        loc='upper left',
        borderaxespad=0.,
        fontsize=8
    )
    figs.append(fig)
    plt.close(fig)

    # Plot a histogram of correlation distributions
    fig = plt.figure(figsize=(8, 6))
    plt.hist(np.abs(correlation_values), bins=100, edgecolor='black')
    plt.title('Histogram of Pairwise Electrode Correlations')
    plt.xlabel('Correlation Coefficient')
    plt.ylabel('Number of Electrode Pairs')
    figs.append(fig)
    plt.close(fig)

    return figs

def ASSR_erp_plot(ERP_c, ERP_n, noised_c_idx, noised_n_idx, roi, comp, baseline=0.2, save=None, vis=False):
    electrodes_to_plot = roi[comp]
    def get_picks_and_data(ERP, electrodes):
        picks = mne.pick_channels(ERP.epochs.info['ch_names'], include=electrodes)
        epochs = ERP.epochs.pick(picks)
        return epochs 
    
    epochs_c = get_picks_and_data(ERP_c, electrodes_to_plot)
    epochs_n = get_picks_and_data(ERP_n, electrodes_to_plot)

    n_epochs_c = len(epochs_c)
    n_epochs_n = len(epochs_n)

    sf = int(epochs_c.info['sfreq'])
    fig_erp = plt.figure(figsize=(16, 12))
    gs_erp = GridSpec(4, 2, figure=fig_erp, hspace=0.7)
    fig_erp.suptitle(f'Event-Related Potencial (ERP), \nCongruent epochs:{n_epochs_c}/{n_epochs_c+len(noised_c_idx)}, Incongruent epochs:{n_epochs_n}/{n_epochs_n+len(noised_n_idx)}')
    t = epochs_c.times
    for i, electrode in enumerate(electrodes_to_plot):
        row, col = divmod(i, 2)
        ax_erp = fig_erp.add_subplot(gs_erp[row, col])
        data_c = epochs_c.get_data(picks=electrode, verbose=False)*1e6
        data_n = epochs_n.get_data(picks=electrode, verbose=False)*1e6
        mean_data_c = np.mean(data_c, axis=0).squeeze()
        mean_data_n = np.mean(data_n, axis=0).squeeze()
        SE_c = np.std(data_c, axis=0, ddof=1).squeeze()/ np.sqrt(data_c.shape[0])
        SE_n = np.std(data_n, axis=0, ddof=1).squeeze()/ np.sqrt(data_n.shape[0])
        

        #  ERP
        ax_erp.plot(t, mean_data_c, label='con', color='blue')
        ax_erp.plot(t, mean_data_n, label='incon', color='red')
        ax_erp.fill_between(t, mean_data_c - SE_c, mean_data_c + SE_c, color='blue', alpha=0.3)
        ax_erp.fill_between(t, mean_data_n - SE_n, mean_data_n + SE_n, color='red', alpha=0.3)
        ax_erp.set_ylim([-20, 20])
        ax_erp.set_title(f'{electrode} electrode',  fontsize=10)
        ax_erp.set_xlabel('Time, sec', fontsize=10)
        ax_erp.set_ylabel('Amplitude, $\mu$V', fontsize=10)
        ax_erp.axhline(0, color='black', linewidth=1)
        ax_erp.axvline(0, color='black', linewidth=1)
        ax_erp.grid(True)
        ax_erp.legend(fontsize=8)
        ax_erp.set_box_aspect(0.2)  # ,      

    #     ERP
    electrode_c_data = [epochs_c.average(picks=electrode, method='mean').data *1e6 for electrode in electrodes_to_plot]
    electrode_n_data = [epochs_n.average(picks=electrode, method='mean').data *1e6 for electrode in electrodes_to_plot]
    mean_data_c_all = np.mean(electrode_c_data, axis=0).squeeze()
    mean_data_n_all = np.mean(electrode_n_data, axis=0).squeeze()

    #    ERP   
    ax_erp_avg = fig_erp.add_subplot(gs_erp[3, 0])
    ax_erp_avg.plot(t, mean_data_c_all, label='con', color='blue', linestyle='--')
    ax_erp_avg.plot(t, mean_data_n_all, label='incon', color='red', linestyle='--')
    ax_erp_avg.set_ylim([-20, 20])
    ax_erp_avg.set_title(f'Average over epochs-averaged electrodes', fontsize=10)
    ax_erp_avg.set_xlabel('Time, sec', fontsize=10)
    ax_erp_avg.set_ylabel('Amplitude, $\mu$V', fontsize=10)
    ax_erp_avg.axhline(0, color='black', linewidth=1)
    ax_erp_avg.axvline(0, color='black', linewidth=1)
    ax_erp_avg.grid(True)
    ax_erp_avg.legend(fontsize=8)
    ax_erp_avg.set_box_aspect(0.2)

    #     
    epochs_c_data = np.mean(epochs_c.get_data(picks=electrodes_to_plot, verbose=False), axis=1)*1e6
    epochs_n_data = np.mean(epochs_n.get_data(picks=electrodes_to_plot, verbose=False), axis=1)*1e6
    mean_data_c_all_epochs = np.mean(epochs_c_data, axis=0)
    mean_data_n_all_epochs = np.mean(epochs_n_data, axis=0)
    SD_c_all_epochs = np.std(epochs_c_data, axis=0, ddof=1)/ np.sqrt(epochs_c_data.shape[0])
    SD_n_all_epochs = np.std(epochs_n_data, axis=0, ddof=1)/ np.sqrt(epochs_n_data.shape[0])
    t_epochs = np.arange(-baseline, len(mean_data_c_all_epochs) / sf - baseline, 1 / sf)

    #    ERP   
    ax_erp_avg_epochs = fig_erp.add_subplot(gs_erp[3, 1])
    ax_erp_avg_epochs.plot(t_epochs, mean_data_c_all_epochs, label='con', color='blue', linestyle='--')
    ax_erp_avg_epochs.plot(t_epochs, mean_data_n_all_epochs, label='incon', color='red', linestyle='--')
    ax_erp_avg_epochs.fill_between(t_epochs, mean_data_c_all_epochs - SD_c_all_epochs, mean_data_c_all_epochs + SD_c_all_epochs, color='blue', alpha=0.3)
    ax_erp_avg_epochs.fill_between(t_epochs, mean_data_n_all_epochs - SD_n_all_epochs, mean_data_n_all_epochs + SD_n_all_epochs, color='red', alpha=0.3)
    ax_erp_avg_epochs.set_ylim([-20, 20])
    ax_erp_avg_epochs.set_title(f'Average over electrodes and epochs', fontsize=10)
    ax_erp_avg_epochs.set_xlabel('Time, sec', fontsize=10)
    ax_erp_avg_epochs.set_ylabel('Amplitude, $\mu$V', fontsize=10)
    ax_erp_avg_epochs.axhline(0, color='black', linewidth=1)
    ax_erp_avg_epochs.axvline(0, color='black', linewidth=1)
    ax_erp_avg_epochs.grid(True)
    ax_erp_avg_epochs.legend(fontsize=8)
    ax_erp_avg_epochs.set_box_aspect(0.2)
    plt.show()
    if not vis:
        plt.close(fig_erp)
    if save:
        fig_erp.savefig(save)

def ASSR_psd_plot(ERP_c, ERP_n, noised_c_idx, noised_n_idx, roi, comp, save=None, vis=False):
    electrodes_to_plot = roi[comp]
    def get_picks_and_data(ERP, electrodes):
        picks = mne.pick_channels(ERP.epochs.info['ch_names'], include=electrodes)
        epochs = ERP.epochs.pick(picks)
        return epochs 
    
    epochs_c = get_picks_and_data(ERP_c, electrodes_to_plot)
    epochs_n = get_picks_and_data(ERP_n, electrodes_to_plot)

    sf = int(epochs_c.info['sfreq'])
    fig_psd = plt.figure(figsize=(16, 12))  #      4:2
    gs_psd = GridSpec(4, 2, figure=fig_psd, hspace=0.7)
    fig_psd.suptitle('PSD')
    
    for i, electrode in enumerate(electrodes_to_plot):
        row, col = divmod(i, 2)
        ax_psd = fig_psd.add_subplot(gs_psd[row, col])

        mean_data_c = np.mean(epochs_c.average(picks=electrode, method='mean').data, axis=0)
        mean_data_n = np.mean(epochs_n.average(picks=electrode, method='mean').data, axis=0)
        
        #    
        psd_c, freqs_c = mne.time_frequency.psd_array_welch(mean_data_c, sfreq=sf, fmin=0, fmax=40, n_fft=sf, verbose=False)
        psd_n, freqs_n = mne.time_frequency.psd_array_welch(mean_data_n, sfreq=sf, fmin=0, fmax=40, n_fft=sf, verbose=False)

        ax_psd.plot(freqs_c, 10 * np.log10(psd_c), label='con', color='blue')
        ax_psd.plot(freqs_n, 10 * np.log10(psd_n), label='n-con', color='red')
        ax_psd.set_title(f'{electrode} electrode', fontsize=10)
        ax_psd.set_xlabel('Frequency (Hz)', fontsize=10)
        ax_psd.set_ylabel('Power (dB)', fontsize=10)
        #ax_psd.axhline(-100, color='black', linewidth=1)
        ax_psd.axvline(0, color='black', linewidth=1)
        ax_psd.grid(True)
        ax_psd.legend(fontsize=8)
        ax_psd.set_box_aspect(0.2)  # ,      

    electrode_c_data = [epochs_c.average(picks=electrode, method='mean').data *1e6 for electrode in electrodes_to_plot]
    electrode_n_data = [epochs_n.average(picks=electrode, method='mean').data *1e6 for electrode in electrodes_to_plot]
    mean_data_c_all = np.mean(electrode_c_data, axis=0).squeeze()
    mean_data_n_all = np.mean(electrode_n_data, axis=0).squeeze()

    #    PSD
    psd_c_all, freqs_c_all = mne.time_frequency.psd_array_welch(mean_data_c_all, sfreq=sf, fmin=0, fmax=40, n_fft=sf, verbose=False)
    psd_n_all, freqs_n_all = mne.time_frequency.psd_array_welch(mean_data_n_all, sfreq=sf, fmin=0, fmax=40, n_fft=sf, verbose=False)

    ax_psd_avg = fig_psd.add_subplot(gs_psd[3, :])
    ax_psd_avg.plot(freqs_c_all, 10 * np.log10(psd_c_all), label='con', color='blue', linestyle='--')
    ax_psd_avg.plot(freqs_n_all, 10 * np.log10(psd_n_all), label='n-con', color='red', linestyle='--')
    ax_psd_avg.set_title(f'Average over electrodes', fontsize=10)
    ax_psd_avg.set_xlabel('Frequency (Hz)', fontsize=10)
    ax_psd_avg.set_ylabel('Power (dB)', fontsize=10)
    #ax_psd_avg.axhline(-100, color='black', linewidth=1)
    ax_psd_avg.axvline(0, color='black', linewidth=1)
    ax_psd_avg.grid(True)
    ax_psd_avg.legend(fontsize=8)
    ax_psd_avg.set_box_aspect(0.1)
    plt.show()
    if not vis:
        plt.close(fig_psd)
    if save:
        fig_psd.savefig(save)

def snr_plot(snr_matrix, ch_names):
    snr_fig, ax = plt.subplots(figsize=(8, 6))
    #    
    ep_ind = np.arange(0, snr_matrix.shape[0], 1)
    ch_ind = np.arange(0, snr_matrix.shape[1], 1)
    #   
    xticklabels = np.array(ep_ind)[::max(1, len(ep_ind) // 10)]
    yticklabels = np.array(ch_names)[ch_ind]
    
    #   
    sns.heatmap(snr_matrix.T, cmap='magma', xticklabels=xticklabels, yticklabels=yticklabels,
                vmin=-30, vmax=-10, ax=ax)
    
    #    
    ax.set_xticks(np.arange(len(xticklabels)))
    ax.set_xticklabels(xticklabels)
    ax.set_title('SNR Heatmap')
    ax.set_xlabel('Epochs')
    ax.set_ylabel('Channels')
    ax.tick_params(axis='y', labelsize=8)
    plt.close(snr_fig)

    snr_values = snr_matrix.flatten()
    #  
    hist_fig, ax = plt.subplots(figsize=(8, 6))
    sns.histplot(snr_values, bins=30, color='blue', edgecolor='black', ax=ax, kde=True, stat='density')
    #  
    ax.set_title('SNR Value Frequency Histogram')
    ax.set_xlabel('SNR Value')
    ax.set_ylabel('Probability')
    ax.grid(axis='y', alpha=0.75)
    ax.set_xlim(-30, 0)
    #  
    plt.close(hist_fig)
    return snr_fig, hist_fig

def compared_snr(snr_matrices, names):
    #    
    kde_fig, ax = plt.subplots(figsize=(10, 6))
    colors = sns.color_palette("husl", len(snr_matrices))

    for snr_matrix, name, color in zip(snr_matrices, names, colors):
        snr_values = snr_matrix.flatten()
        mean_snr = np.median(snr_values)
        label_with_mean = f"{name} (Median: {mean_snr:.2f})"
        kde = sns.kdeplot(snr_values, ax=ax, color=color, fill=True, bw_adjust=0.5, label=label_with_mean, alpha=0.5)
        
        color = kde.collections[-1].get_facecolor()[0]  #    RGBA
        ax.axvline(mean_snr, linestyle='--', color=color, linewidth=1.5)

    #  
    ax.set_title('SNR KDE')
    ax.set_xlabel('SNR Value')
    ax.set_ylabel('Density')
    ax.grid(axis='y', alpha=0.75)
    ax.set_xlim(-35, 0)
    ax.set_ylim(0, 0.15)
    ax.grid()
    ax.legend(title='SNR Matrices', loc='upper right')
    plt.close(kde_fig)
    return kde_fig

def bridge_save(raw, correlation_matrix, bridged_idx, ed_matrix, saving_dir=None, vis=False):
    channels = raw.ch_names
    bridge_fig = mne.viz.plot_bridged_electrodes(raw.info, bridged_idx, ed_matrix)
    fig_corr = plt.figure(figsize=(8, 6))
    ax = sns.heatmap(correlation_matrix, cmap='viridis', xticklabels=channels, yticklabels=channels) #np.mean(ed_matrix, axis=0)
    ax.set_xticklabels(channels, rotation=90, fontsize=7)
    ax.set_yticklabels(channels, rotation=0, fontsize=7)
    plt.title('Correlation of Channel Connections (R)')
    plt.xlabel('Channels')
    plt.ylabel('Channels')
    if saving_dir:
        plt.savefig(os.path.join(saving_dir, f'orrelation_electrodes.png'))
        bridge_fig.savefig(os.path.join(saving_dir, f'bridged_electrodes.png'))
    if not vis:
        plt.close(fig_corr)
        plt.close(bridge_fig)

def components_save(raw, ica, exclude_idx, labels, probas, saving_dir=None, vis=False):
    duration = raw.times[-1]*1000
    start_time = 0 
    stop_time = min(10*1000, duration)  
    ica_fig = ica.plot_components(title='ICA Components', show=vis)
    if isinstance(raw, mne.io.Raw) or isinstance(raw,mne.io.brainvision.brainvision.RawBrainVision):
        overlay_fig = ica.plot_overlay(raw, title='Overlay', start=start_time, stop=stop_time, 
                                    exclude=exclude_idx, show=vis)
    else:
        overlay_fig = ica.plot_overlay(raw.average(), title='Overlay', start=start_time, stop=stop_time, 
                                    exclude=exclude_idx, show=vis)
    if saving_dir:
        ica_fig.savefig(os.path.join(saving_dir, 'ica_components.png'))
        overlay_fig.savefig(os.path.join(saving_dir, 'overlay.png'))
    if not vis:
        plt.close(ica_fig)
        plt.close(overlay_fig)
    for idx in exclude_idx:
        ica_exclude_fig = ica.plot_properties(raw, picks=idx, show=vis, verbose=False)[0]
        if saving_dir:
            if labels is None:
                ica_exclude_fig.savefig(os.path.join(saving_dir, f'ica_exclude_components_{idx}_label:hand.png'))
            else:
                ica_exclude_fig.savefig(os.path.join(saving_dir, f'ica_exclude_components_{idx}_{labels[idx]}_{probas[idx]}.png'))
        if not vis:
            plt.close(ica_exclude_fig)

def event_saver(proc_counter, mind_counter, saving_dir=None, vis=False):
    df = pd.read_csv('tools/ASSR.csv', index_col=None)
    combined_dict = {**proc_counter, **mind_counter}
    combined_dict = {key.replace('Stimulus/', ''): value for key, value in combined_dict.items()}
    df.columns = df.columns.str.replace(' ', '')
    new_row = {column: combined_dict[column] for column in df.columns}
    new_row_df = pd.DataFrame([new_row])
    df = pd.concat([df, new_row_df], ignore_index=True)
    target_row = df.iloc[0].copy()
    target_row_numeric = pd.to_numeric(target_row, errors='coerce')
    new_row_numeric = pd.to_numeric(pd.Series(new_row), errors='coerce')
    difference_row = target_row_numeric - new_row_numeric
    difference_row_df = pd.DataFrame([difference_row])
    df = pd.concat([df, difference_row_df], ignore_index=True)

    df.insert(0, 'Cond', ['Target', 'Real', 'Diff'])
    if saving_dir:
        df.to_csv(os.path.join(saving_dir, 'stimulus_counts.csv'), index=False)
    fig_table, ax = plt.subplots(figsize=(12, 2))  
    ax.axis('off')
    table = ax.table(cellText=df.values, colLabels=df.columns, cellLoc='center', loc='center')
    table.auto_set_font_size(False)
    table.set_fontsize(10)
    table.scale(1.2, 1.2)
    plt.title('  ', fontsize=14, pad=1)
    if saving_dir:
        fig_table.savefig(os.path.join(saving_dir, 'stimulus_counts.png'), bbox_inches='tight', pad_inches=0.1)
    if not vis:
        plt.close(fig_table)

def tfr_plot(tfr, itc, picks, fmin, fmax, vmin=None, vmax=None, baseline=(None, 0), combine='mean', 
             baseline_mode='logratio', saving_dir=None, vis=False, 
             tfr_title='Event Related Spectral Perturbation (ERSP)',
             itc_title='Inter-Trial Coherence (ITC)'):  
    tfr_fig = tfr.plot(title=tfr_title, fmin=fmin, fmax=fmax, vlim=(vmin,vmax), 
             picks=picks, combine=combine, baseline=baseline, mode=baseline_mode)[0] 
    itc_fig = itc.plot(title=itc_title, fmin=fmin, fmax=fmax, 
             picks=picks, combine=combine, baseline=baseline, mode=baseline_mode)[0]
    if not vis:
        plt.close(tfr_fig)
        plt.close(itc_fig)
    if saving_dir:
        tfr_fig.savefig(os.path.join(saving_dir, f'ERSP.png'))
        itc_fig.savefig(os.path.join(saving_dir, f'ITC.png'))

def plot_noise_data(raw, noisy_channel_flags, n_channels, threshold_noise, integral_curve):
    #    GridSpec   
    fig = plt.figure(figsize=(18, 8))
    gs = GridSpec(1, 3, width_ratios=[5, 1, 3], figure=fig)

    #    
    ax0 = fig.add_subplot(gs[0, 0])
    sns.heatmap(noisy_channel_flags, vmin=0, vmax=1, cmap='viridis', 
                cbar_kws={'label': '  PSD'}, ax=ax0)
    ax0.set_xlabel('')
    ax0.set_ylabel('')
    ax0.set_title('   PSD    ')
    ax0.set_yticks(np.arange(n_channels) + 0.5)
    ax0.set_yticklabels(raw.ch_names, rotation=0)
    
    #   
    ax1 = fig.add_subplot(gs[0, 1], sharey=ax0)  #  sharey   Y  ax0  ax1
    ax1.plot(integral_curve, np.arange(n_channels)+0.5, color='blue', label='  ')
    ax1.axvline(threshold_noise, color='red', linestyle='--', label=f' - {threshold_noise}')
    ax1.set_xlim((0, 1))
    ax1.set_xlabel(' ')
    ax1.set_title(' ')
    ax1.legend(loc='right', fontsize='small')
    ax1.grid(True)

    #  
    ax2 = fig.add_subplot(gs[0, 2])  #    
    pos = np.array([raw.info['chs'][i]['loc'][:2] for i in range(raw.info['nchan'])])  # X  Y 
    mne.viz.plot_topomap(
        integral_curve, pos, axes=ax2, show=False, cmap='RdBu_r', vlim=(0,1),
        names=raw.ch_names
    )
    ax2.set_title('  ')
    
    plt.tight_layout()
    plt.show()

def plot_topomap(coherence, band_name, info, title):
    #      
    coherence_mean = coherence.mean(axis=1)
    #    
    mask = np.zeros((len(info['ch_names']), len(info['ch_names'])), dtype=bool)
    for i in range(len(info['ch_names'])):
        for j in range(i+1, len(info['ch_names'])):
            mask[i, j] = True

    #   
    mne.viz.plot_topomap(coherence_mean, info, mask=mask, show=False)
    plt.title(title)
    plt.show()

def char_plot(spec_con, epochs, method, mode, cwt_freqs):
    con = spec_con.get_data()
    if np.iscomplexobj(con):
        con = np.abs(con)
    con_mean = con.mean(axis=0)
    times = epochs.times
    #  
    plt.figure(figsize=(12, 6))
    plt.imshow(con_mean, aspect=1/40, cmap='inferno', extent=[times[0], times[-1], cwt_freqs[0], cwt_freqs[-1]], origin='lower') #vmin=0, vmax=1
    plt.colorbar(label=f' {method} {mode}')
    plt.xlabel('Time (s)')
    plt.ylabel('Frequency (Hz)')
    plt.title(f'  {method}   CWT Morlet')
    plt.show()
    plt.savefig(f'{method} {mode}.png')

def plot_circular_connectivity(con, epochs, threshold=0.5, title='Circular Connectivity Diagram'):
    con = np.mean(np.abs(con), axis=2)  #   
    con[con < threshold] = 0  #    
    node_names = epochs.ch_names
    mne.viz.plot_connectivity_circle(con, node_names, title=title, vmin=0, vmax=1, colormap='inferno', show=True)



# Copyright 2025 Sear Gamemode
import mne 
import os
import numpy as np

from tqdm import tqdm
import matplotlib.pyplot as plt
from collections import defaultdict
from mne.preprocessing import compute_bridged_electrodes
from itertools import chain
from scipy.sparse import csr_matrix
from scipy.sparse.csgraph import connected_components

from autoreject import Ransac 
from joblib import Parallel, delayed
from .savers import bridge_save, event_saver, plot_clusters
from .metrics import isolation_forest, check_volt_of_epochs 
from .scenarious import get_riti_seq, verificate_events, is_subsequence
from .montages import create_custom_montage, read_elc, align_head

def compute_bad_epochs(epochs, snr_matrix, roi_channels=None, thr_auto=True):
    n_trials, n_channels, n_times = epochs._data.shape
    ch_names = epochs.ch_names 
    
    if thr_auto:
        threshold_ep=np.percentile(snr_matrix, 20, axis=(0, 1))
        threshold_ch=np.percentile(snr_matrix, 0.5, axis=(0, 1))
        volt_max=dict(eeg=np.percentile(np.abs(epochs._data), 99.999, axis=(0, 1, 2)))
        volt_min=dict(eeg=np.percentile(np.abs(epochs._data), 10, axis=(0, 1, 2)))
    else:
        threshold_ep=-22
        threshold_ch=-30
        volt_max = dict(eeg=130e-6)
        volt_min = dict(eeg=1e-6)

    #--------------SNR---------------#
    rej_dict = {}
    
    rej_dict['SNR_over_channels'] = np.where(np.mean(snr_matrix, axis=1) < threshold_ep)[0].tolist() #   ,      
    rej_dict['SNR_channel'] = []
    for epoch_idx in range(n_trials):
        channel_names_above_threshold = [ch_names[ch] for ch in np.where(snr_matrix[epoch_idx] < threshold_ch)[0]] #     
        if channel_names_above_threshold:
            rej_dict['SNR_channel'].append((epoch_idx, channel_names_above_threshold)) #         

    #-------ISOLATION_FOREST---------#
    rej_dict['Isol_forest_all_chns'] = isolation_forest(epochs.copy(), mode='ep')
    if roi_channels:
        rej_dict['Isol_forest_tar_chns'] = isolation_forest(epochs.copy().pick(roi_channels), mode='ep')
    #------------VOLT----------------#
    rej_dict['Volt_max'] = check_volt_of_epochs(epochs, reject=volt_max, flat=None)
    rej_dict['Volt_flat'] = check_volt_of_epochs(epochs, reject=None, flat=volt_min)

    #--------------FINAL-------------#
    snr_all_channels_indices = set(rej_dict['SNR_over_channels'])
    snr_channel_indices = set(epoch for epoch, _ in rej_dict['SNR_channel'])
    volt_max_indices = set(rej_dict['Volt_max'])
    volt_min_indices = set(rej_dict['Volt_flat'])
    isol_forest_all_indices = set(rej_dict['Isol_forest_all_chns'])

    combined_indices = snr_all_channels_indices | snr_channel_indices | volt_max_indices | isol_forest_all_indices | volt_min_indices
    if roi_channels:
        isol_forest_tar_indices = set(rej_dict['Isol_forest_tar_chns'])
        combined_indices = combined_indices | isol_forest_tar_indices
        
    rej_dict['FINAL'] = sorted(list(combined_indices))
    rej_dict['Percentage_removed_trials'] = len(rej_dict['FINAL'])/epochs._data.shape[0]*100
    #for key in rej_dict.keys():
    #    print(f'rej_dict[{key}] = {rej_dict[key]}')
    return rej_dict

def set_montage(raw, montage, elc_file, mode, threshold, verbose=False, interpolate=None, vis=None):
    if montage=='waveguard64':
        montage = create_custom_montage(montage)
    elif montage == 'personal':
        ch_dict, nasion, lpa, rpa, hsp = read_elc(elc_file)
        if interpolate:
            ch_dict, nasion, lpa, rpa, hsp = align_head(ch_dict, nasion, np.array(lpa), np.array(rpa), np.array(hsp), standard='waveguard64', 
                                                    mode=mode, threshold=threshold)
        montage = mne.channels.make_dig_montage(ch_pos=ch_dict, nasion=nasion, lpa=lpa, rpa=rpa, hsp=hsp, coord_frame='head')
    else:
        montage = mne.channels.make_standard_montage(montage)
    
    raw.set_montage(montage, verbose=False)

    if vis:
        fig = montage.plot(show_names=True, kind='3d')
        for ax in fig.get_axes():
            if ax.name == '3d':
                ax.set_xlim([-0.1, 0.1])
                ax.set_ylim([-0.1, 0.1])
                ax.set_zlim([-0.1, 0.1])
    return raw

def detect_bad_channels(raw, method):
    bad_channels = []
    scores = []
    electrodesD = {}
    flat_chans, noisy_channels = get_flat_channels(raw)
    clusters, bridge_figs = search_bridge_cluster(raw, method='corr')
    bridged_electrodes = list(chain.from_iterable(clusters))
    
    if method == 'ransac':
        bad_channels, scores, noised_fig = DNC_ransac(raw)
    elif method == 'ed':
        bad_channels, scores, noised_fig = DNC_electrical_distance(raw)
    elif method == 'corr':
        bad_channels, scores, noised_fig = DNC_corr(raw)
    elif method in ['auto', 'SN_rate']:
        bad_channels, scores, noised_fig = DNC_SN_rate(raw)
    else:
        raise ValueError(f"Unknown method '{method}'. Please use 'ransac', 'neighbours', 'psd', 'ed', 'corr' or 'auto'")
    
    electrodesD['HighAmp'] = noisy_channels
    electrodesD['LowAmp'] = flat_chans
    electrodesD['Bridged'] = bridged_electrodes
    electrodesD['Noise_Rate'] = bad_channels
    return electrodesD, clusters, bridge_figs, noised_fig


def preprocessing_events(raw, scenarious='RiTi'):
    if scenarious == 'RiTi':
        seq_list = get_riti_seq()
        events, event_id = mne.events_from_annotations(raw, verbose=False)
        reverse_event_id = dict(zip(event_id.values(), event_id.keys()))
        new_events = verificate_events(raw, seq_list)
        filtered_seq = [reverse_event_id[new_events[j, 2]] for j in range(len(new_events))]
        subseq, match_idxs, tr = is_subsequence(filtered_seq, seq_list)
        repo_data = {
                    'Initial_Events': len(events),
                    'Filtered_Events': len(new_events),
                    'Expected_Events': len(seq_list),
                    'SubSeq_Flag': subseq, 
                    'Match_Indexes': match_idxs,
                    'Transitions': tr
                    }
        return new_events, reverse_event_id, repo_data
    else:
        print(' !')
        return None
    

def get_flat_channels(raw):
    data = raw.get_data()
    channel_names = raw.ch_names
    threshold_min_amplitude = 3e-6
    threshold_max_amplitude = 300e-6
    threshold_length = 0.5
    empty_channels = [ch for idx, ch in enumerate(channel_names) if np.mean(np.abs(data[idx]) < threshold_min_amplitude) > threshold_length]
    empty_scores = [np.mean(np.abs(data[idx] - data[idx].mean(keepdims=True)) < threshold_min_amplitude) for idx, ch in enumerate(channel_names) if np.mean(np.abs(data[idx]) < threshold_min_amplitude) > threshold_length]
    noisy_channels = [ch for idx, ch in enumerate(channel_names) if np.mean(np.abs(data[idx]) > threshold_max_amplitude) > threshold_length]
    noisy_scores = [np.mean(np.abs(data[idx]) > threshold_max_amplitude) for idx, ch in enumerate(channel_names) if np.mean(np.abs(data[idx]) > threshold_max_amplitude) > threshold_length]
    return empty_channels, noisy_channels


def search_bridge_cluster(raw, method='corr', threshold_bridge=0.99, threshold_cluster=0.98):
    """
    Find clusters of bridged electrodes based on correlation or predefined methods.

    Parameters:
    raw : mne.io.Raw
        Raw EEG data.
    method : str
        Method to detect bridges ('corr', 'auto', or 'ed').
    threshold_bridge : float
        Threshold for identifying bridged electrodes.
    threshold_cluster : float
        Threshold for identifying clusters of bridged electrodes.

    Returns:
    clusters : list
        List of clusters of bridged electrodes.
    figs : list
        Figures visualizing the clusters.
    """
    # Extract data and initialize variables
    data = raw.get_data()
    ch_names = raw.ch_names
    ch_name_to_idx = {ch: idx for idx, ch in enumerate(ch_names)}
    adjacency_list = defaultdict(list)

    # Step 1: Detect bridged electrodes
    if method == 'corr':
        correlation_matrix = np.corrcoef(data)
        for i in range(len(ch_names)):
            for j in range(i + 1, len(ch_names)):
                if correlation_matrix[i, j] >= threshold_bridge:
                    adjacency_list[ch_names[i]].append(ch_names[j])
                    adjacency_list[ch_names[j]].append(ch_names[i])

    elif method in ['auto', 'ed']:
        bridged_idx, _ = compute_bridged_electrodes(raw, verbose=False)
        for ch1, ch2 in bridged_idx:
            adjacency_list[ch_names[ch1]].append(ch_names[ch2])
            adjacency_list[ch_names[ch2]].append(ch_names[ch1])

    # Step 2: Depth-first search (DFS) for clustering
    def dfs(electrode, cluster, visited):
        idx_electrode = ch_name_to_idx[electrode]

        # Check if the electrode can be added to the cluster
        for existing_electrode in cluster:
            idx_existing = ch_name_to_idx[existing_electrode]
            if correlation_matrix[idx_electrode, idx_existing] < threshold_cluster:
                return

        cluster.append(electrode)
        visited.add(electrode)

        for neighbor in adjacency_list[electrode]:
            if neighbor not in visited:
                dfs(neighbor, cluster, visited)

    # Step 3: Find clusters
    clusters = []
    visited = set()
    for electrode in ch_names:
        if electrode not in visited and electrode in adjacency_list:
            cluster = []
            dfs(electrode, cluster, visited)
            clusters.append(cluster)

    # Step 4: Analyze clusters (if method is 'corr')
    if method == 'corr':
        clusters_avg, clusters_max, clusters_min = [], [], []
        for cluster in clusters:
            cluster_indices = [ch_name_to_idx[ch] for ch in cluster]
            cluster_corr_matrix = correlation_matrix[np.ix_(cluster_indices, cluster_indices)]

            # Extract upper triangle values of the correlation matrix
            triu_indices = np.triu_indices_from(cluster_corr_matrix, k=1)
            cluster_corr_values = cluster_corr_matrix[triu_indices]

            if len(cluster_corr_values) > 0:
                clusters_avg.append(np.mean(cluster_corr_values))
                clusters_max.append(np.max(cluster_corr_values))
                clusters_min.append(np.min(cluster_corr_values))

    # Step 5: Generate plots
    ch_pos = np.array([raw.info['chs'][i]['loc'][:2] for i in range(raw.info['nchan'])])
    figs = plot_clusters(clusters, clusters_avg, clusters_max, clusters_min, ch_names, ch_pos, correlation_matrix)

    return clusters, figs


def search_bridge_cluster_with_times(raw, method='corr', threshold_bridge=0.99, threshold_cluster=0.98, window_size=2.0, overlap=0.5):
    """
    Find clusters of bridged electrodes and determine the time of bridging using a sliding window approach.

    Parameters:
    raw : mne.io.Raw
        Raw EEG data.
    method : str
        Method to detect bridges ('corr', 'auto', or 'ed').
    threshold_bridge : float
        Threshold for identifying bridged electrodes.
    threshold_cluster : float
        Threshold for identifying clusters of bridged electrodes.
    window_size : float
        Size of the sliding window in seconds.
    overlap : float
        Overlap between consecutive windows (0 to 1).

    Returns:
    clusters : list
        List of clusters of bridged electrodes for the selected window.
    bridge_times : dict
        Dictionary where keys are electrode pairs and values are lists of window indices where bridging occurred.
    figs : list
        Figures visualizing the clusters.
    """
    data = raw.get_data()
    sfreq = raw.info['sfreq']
    ch_names = raw.ch_names
    ch_name_to_idx = {ch: idx for idx, ch in enumerate(ch_names)}
    window_samples = int(window_size * sfreq)
    step_samples = int(window_samples * (1 - overlap))
    n_channels = len(ch_names)

    clusters = []
    bridge_times = defaultdict(list)
    clusters_avg = []
    clusters_max = []
    clusters_min = []
    correlation_values = []

    for start in range(0, data.shape[1] - window_samples + 1, step_samples):
        end = start + window_samples
        window_data = data[:, start:end]

        if method == 'corr':
            # Step 1: Compute correlation matrix
            correlation_matrix = np.corrcoef(window_data)
            np.fill_diagonal(correlation_matrix, 0)  # Remove self-loops

            # Step 2: Create adjacency matrix
            adjacency_matrix = (correlation_matrix >= threshold_bridge).astype(int)

            # Step 3: Find connected components (clusters)
            sparse_adj_matrix = csr_matrix(adjacency_matrix)
            n_components, labels = connected_components(csgraph=sparse_adj_matrix, directed=False)

            # Map components to clusters
            window_clusters = [
                [ch_names[i] for i in range(n_channels) if labels[i] == component]
                for component in range(n_components)
            ]

            # Combine all clusters into a single list for `plot_clusters`
            clusters.extend(window_clusters)

            # Record bridge times
            for i in range(n_channels):
                for j in range(i + 1, n_channels):
                    if adjacency_matrix[i, j]:
                        bridge_times[(ch_names[i], ch_names[j])].append(start // step_samples)

            # Step 4: Calculate cluster statistics
            for cluster in window_clusters:
                cluster_indices = [ch_name_to_idx[ch] for ch in cluster]
                cluster_corrs = correlation_matrix[np.ix_(cluster_indices, cluster_indices)]
                triu_indices = np.triu_indices_from(cluster_corrs, k=1)
                cluster_corr_values = cluster_corrs[triu_indices]

                if len(cluster_corr_values) > 0:
                    clusters_avg.append(np.mean(cluster_corr_values))
                    clusters_max.append(np.max(cluster_corr_values))
                    clusters_min.append(np.min(cluster_corr_values))

    # Step 5: Visualize clusters
    ch_pos = np.array([raw.info['chs'][i]['loc'][:2] for i in range(n_channels)])
    figs = plot_clusters(clusters, clusters_avg, clusters_max, clusters_min, ch_names, ch_pos, correlation_values)

    return clusters, bridge_times, figs


def event_check(raw, mind_stimulus, proc_stimulus, saving_dir=None, vis=False):
    proc_count = 0
    mind_count = 0
    other_count = 0
    events, event_id = mne.events_from_annotations(raw, verbose=False)
    rev_event_id = dict(list(zip(event_id.values(), event_id.keys())))
    mind_counter = {stimulus: 0 for stimulus in mind_stimulus}
    proc_counter = {stimulus: 0 for stimulus in proc_stimulus}
    strange_list = set()
    for event in events:
        if rev_event_id[event[2]] in proc_counter.keys():
            for key in proc_counter.keys():
                if rev_event_id[event[2]] == key:
                    proc_counter[key]+=1
                    proc_count += 1
        elif rev_event_id[event[2]] in mind_counter.keys():
            for key in mind_counter.keys():
                if rev_event_id[event[2]] == key:
                    mind_counter[key]+=1
                    mind_count += 1
        else:
            strange_list.add(rev_event_id[event[2]])
            other_count += 1
    event_saver(proc_counter, mind_counter, saving_dir, vis)

def bridging_test(raw, saving_dir=None, vis=False):
    raw = raw.copy()
    bridged_idx, ed_matrix = compute_bridged_electrodes(raw, verbose=False)
    correlation_matrix = np.corrcoef(raw.get_data())
    bridge_save(raw, correlation_matrix, bridged_idx, ed_matrix, saving_dir, vis=vis)
    return bridged_idx

def find_adj_neighbors(raw, ch_name):
    """       ch_name"""
    montage = raw.get_montage()
    adjacency, ch_names = mne.channels.find_ch_adjacency(raw.info, 'eeg')
    adjacency = adjacency.toarray()
    idx = raw.ch_names.index(ch_name)
    neighbors_idx = np.where(adjacency[idx])[0]
    neighbors = [ch_names[i] for i in neighbors_idx]
    return neighbors

def calculate_correlations(raw, duration=1, overlap=0.5):
    epochs = mne.make_fixed_length_epochs(raw, duration=duration, overlap=overlap, verbose=False)
    data = epochs.get_data()
    n_trials, n_channels, n_times = epochs.get_data().shape

    adjacency, ch_names = mne.channels.find_ch_adjacency(raw.info, ch_type="eeg")

    neighbors_indices = [np.where(adjacency[i])[0] for i in range(n_channels)]
    correlations = np.zeros((n_trials, n_channels))  #   
    #      
    for i in range(n_trials):
        epoch_data = data[i]  #    
        for j in range(n_channels):
            #     
            neighbor_data = np.mean(epoch_data[neighbors_indices[j]], axis=0)
            channel_data = epoch_data[j]
            #       
            corr = np.corrcoef(channel_data, neighbor_data)[0, 1]
            correlations[i, j] = corr  #  
    return correlations


def DNC_SN_rate(raw, optimized=False):
    def compute_snr(signal, noise):
        snr = 10 * np.log10(np.mean(signal ** 2) / np.mean(noise ** 2))
        return snr
    def snr_db_to_probability(snr_db):
        snr_linear = 10 ** (snr_db / 10)
        probability = snr_linear / (1 + snr_linear)
        return probability
    def plot_topomap(probabilities, raw, high_prob, high_prob_channels):
        """Plot topomap of SNR probabilities."""
        montage = raw.get_montage()
        ch_pos_dict = montage.get_positions()['ch_pos']
        ch_pos = np.array([ch_pos_dict[ch_name][:2] for ch_name in raw.ch_names])
        fig, ax = plt.subplots()
        im, _ = mne.viz.plot_topomap(
            probabilities, ch_pos, axes=ax, show=False,
            names=raw.ch_names, cmap='viridis', vlim=(0.5, 1)
        )
        ax.set_title('SNR Probability Topomap')
        cbar = plt.colorbar(im, ax=ax, shrink=0.8)
        cbar.set_label('SNR Probability Coefficient')
        noisy_text = "Noisy Electrodes:\n" + "\n".join(
            [f"{ch_name}: {prob:.2f}" for ch_name, prob in zip(high_prob_channels, high_prob)]
        )
        fig.text(1.1, 0.5, noisy_text, va='center', ha='left', fontsize=10)
        plt.close(fig)
        return fig
    
    ch_names = raw.ch_names
    original_data = raw.get_data()
    
    def process_channel(ch_idx):
        ch_name = ch_names[ch_idx]
        raw_channel = raw.copy().load_data()
        raw_channel.pick(ch_names)
        raw_channel.info['bads'] = [ch_name]
        noised_data = original_data[ch_idx, :].copy()
        raw_channel.interpolate_bads(reset_bads=False, verbose=False)
        interpolated_data = raw_channel.get_data(picks=[ch_name])[0]
        snr = compute_snr(interpolated_data, noised_data)
        proba = snr_db_to_probability(-snr)
        return ch_name, snr, proba
    
    snr_values = {}
    snr_probabilities = {}
    high_prob_channels = []
    high_prob = []

    if optimized:
        results = Parallel(n_jobs=4)(
            delayed(process_channel)(ch_idx) for ch_idx in range(len(ch_names))
        )
    else:
        results = []
        for ch_idx in range(len(ch_names)):
            result = process_channel(ch_idx)
            results.append(result)
            
    for ch_name, snr, proba in results:
        snr_values[ch_name] = snr
        snr_probabilities[ch_name] = proba
        if proba > 0.8:
            high_prob_channels.append(ch_name)
            high_prob.append(proba)
    
    fig = plot_topomap(list(snr_probabilities.values()), raw, high_prob, high_prob_channels)
    plt.close(fig)
    return high_prob_channels, high_prob, fig

def DNC_corr(raw, threshold_corr=0.65, threshold_perc=5):
    correlations = calculate_correlations(raw)
    
    def plot_topomap(correlations, raw):
        """ topomap   """
        montage = raw.get_montage()
        ch_pos = np.array([raw.info['chs'][i]['loc'][:2] for i in range(raw.info['nchan'])])
        fig, ax = plt.subplots()
        im, _ = mne.viz.plot_topomap(np.abs(correlations), ch_pos, axes=ax, show=False, names=raw.ch_names, cmap='viridis', vlim = (0.5, 1))
        ax.set_title('Correlation topomap')
        cbar = plt.colorbar(im, ax=ax, shrink=0.8)
        cbar.set_label('Correlation Coefficient')
        plt.close(fig)
        return fig
    
    fig = plot_topomap(np.mean(correlations, axis=0), raw)

    result = np.mean(correlations, axis=0)
    p5 = np.percentile(correlations, threshold_perc, axis=(0, 1))
    artefacts_indexes = result < p5

    binary_matrix = correlations < threshold_corr
    probs = binary_matrix.mean(axis=0)
    noised_channels = ((np.array(raw.ch_names)[artefacts_indexes]).tolist())
    score = probs[artefacts_indexes]
    return noised_channels, score, [fig]


def normalization(x):
    q75 = np.percentile(x, 75, axis=(0, 1), keepdims=True)
    q25 = np.percentile(x, 25, axis=(0, 1), keepdims=True)
    iqr = q75 - q25
    x_norm = x/iqr
    return x_norm 



def DNC_electrical_distance(raw, vis=False, thr_std=3):

    def sigmoid(x):
        return 1/(1 + np.exp(-x))
    
    bridged_idx, ed_matrix = compute_bridged_electrodes(raw, verbose=False)
  
    ed_matrix = ed_matrix.copy()
    picks = mne.pick_types(raw.info, eeg=True)
    tril_idx = np.tril_indices(picks.size)
    for epo_idx in range(ed_matrix.shape[0]):
        ed_matrix[epo_idx][tril_idx] = ed_matrix[epo_idx].T[tril_idx]
    channel_names = np.array([raw.ch_names[i] for i in picks])

    ed_matrix = np.nanmin(ed_matrix, axis=1)

    ed_matrix_norm = normalization(ed_matrix)
    elec_dists = np.median(ed_matrix_norm, axis=0)
    
    picks = mne.pick_types(raw.info, eeg=True)
    ch_names = raw.ch_names

    results = (ed_matrix_norm - elec_dists).mean(axis=0)
    probs = sigmoid(results)
    artefact_indices = results > thr_std

    fig, ax = plt.subplots()
    pos = np.array([raw.info['chs'][i]['loc'][:2] for i in range(raw.info['nchan'])])
    mne.viz.plot_topomap(
        probs, pos, show=True, cmap='RdBu_r', vlim=(0, 1.0),
        names=raw.ch_names, axes=ax
    )
    ax.set_title(' ')
    plt.tight_layout()
    plt.close(fig)
    return (channel_names[artefact_indices]).tolist(), probs[artefact_indices], [fig]

def DNC_ransac(raw):  
    ransac = Ransac(verbose=False)
    epochs = mne.make_fixed_length_epochs(raw, duration=10, overlap=0.5, preload=True)
    _ = ransac.fit_transform(epochs)
    probabilities = ransac.bad_log.mean(axis=0)
    bad_channels = ransac.bad_chs_
    scores = probabilities[mne.pick_channels(raw.info['ch_names'], bad_channels)]
    fig, ax = plt.subplots()
    pos = np.array([raw.info['chs'][i]['loc'][:2] for i in range(raw.info['nchan'])])
    mne.viz.plot_topomap(
        probabilities, pos, show=True, cmap='RdBu_r', vlim=(0, 1.0),
        names=raw.ch_names, axes=ax
    )
    ax.set_title(' ')
    plt.tight_layout()
    return bad_channels, scores, fig

def compared_spectrum(inst1, inst2, l_freq=0.5, h_freq=80, fmin=0, fmax=100):
    psd_before = inst1.compute_psd(fmin=fmin, fmax=fmax, 
                            remove_dc=False, verbose=False
    )
    psd_after = inst2.compute_psd(fmin=fmin, fmax=fmax, 
                            remove_dc=False, verbose=False
    )
    if isinstance(inst1, mne.Epochs) and isinstance(inst2, mne.Epochs):
        psd_before = psd_before.average()
        psd_after = psd_after.average()
    #  PSD  dB   
    mean_psd_before = 10*np.log10(psd_before.get_data() * 1e12)
    mean_psd_after = 10*np.log10(psd_after.get_data() * 1e12)

    #  
    freqs = psd_before.freqs

    #    
    bands = {
        'Delta': (max(l_freq, 0.5), 4),
        'Theta': (4, 8),
        'Alpha': (8, 13),
        'Beta': (13, 30),
        'Gamma': (30, min(h_freq, 80))
    }

    #   PSD     
    psd_list = [mean_psd_before, mean_psd_after]
    titles = ['Spectrum before', 'Spectrum after']
    colors = ['b', 'g']

    ymin = np.floor(min(mean_psd_before.min(), mean_psd_after.min()) / 10) * 10
    ymax = np.ceil(max(mean_psd_before.max(), mean_psd_after.max()) / 10) * 10

    #    
    fig, axs = plt.subplots(2, 1, figsize=(14, 10), sharex=False)

    #   PSD    
    for idx, (mean_psd, title, color) in enumerate(zip(psd_list, titles, colors)):
        ax = axs[idx]
        y_mean = np.mean(mean_psd, axis=0)
        y_std = np.std(mean_psd, axis=0)
        
        #  
        ax.plot(freqs, y_mean, color=color)
        ax.fill_between(freqs, y_mean - y_std, y_mean + y_std, color=color, alpha=0.3)
        ax.set_title(title)
        ax.set_ylabel('PSD (dB/V/Hz)')
        ax.grid(True)
        
        #        
        band_powers = {}
        for band_name, (fmin, fmax) in bands.items():
            idx_band = np.logical_and(freqs >= fmin, freqs <= fmax)
            band_power = y_mean[idx_band].mean()
            band_powers[band_name] = band_power
            
            #       
            ax.axvline(fmin, color='red', linestyle='--', linewidth=1)
            ax.axvline(fmax, color='red', linestyle='--', linewidth=1)
            ax.fill_betweenx([ymin, ymax], fmin, fmax, color='grey', alpha=0.05)
            
            #      
            ax.text((fmin + fmax) / 2, ymax - 2, f"{band_name}\n{band_power:.1f} dB",
                    horizontalalignment='center', verticalalignment='top', fontsize=9, zorder=4)
        
        #      Y
        ax.set_ylim(ymin, ymax)
        ax.set_xlim(freqs.min(), freqs.max())
        ax.set_xticks(np.arange(freqs.min(), freqs.max()+1, 2))
        ax.set_yticks(np.arange(ymin, ymax+1, 5))
        ax.set_xlabel('Frequency (Hz)')

    #     
    plt.tight_layout()
    plt.close(fig)
    return fig



# Copyright 2025 Sear Gamemode
import mne 

import numpy as np 


def calculate_tfr(epochs, method, fmin=2, fmax=40, n_coef=3.5, average=True, return_itc=False):
    sfreq=epochs.info['sfreq']
    num=int(n_coef*(fmax-fmin))
    freqs = np.logspace(*np.log10([fmin, fmax]), num=num)  
    n_cycles = (freqs*np.pi*(len(epochs.times)-1))/(5*sfreq)
    if return_itc:
        tfr, itc = epochs.compute_tfr(method=method, freqs=freqs, n_cycles=n_cycles, average=average, 
                                      return_itc=True, n_jobs=7, use_fft=False, zero_mean=True)
        return tfr, itc, freqs
    else:
        tfr = epochs.compute_tfr(method=method, freqs=freqs, n_cycles=n_cycles, average=average, 
                                 return_itc=False, n_jobs=7, use_fft=False, zero_mean=True)
        return tfr, freqs



# Copyright 2025 Sear Gamemode
import emd
import mne

import numpy as np 

from scipy.signal import stft, istft
from sklearn.decomposition import PCA
from itertools import combinations


def info_suppression(epochs, n_components=10):
    n_epochs, n_channels, n_times = epochs.get_data().shape
    reshaped_data = epochs.get_data().reshape(-1, n_times)
    print(reshaped_data.shape, epochs.info['sfreq'])
    #  
    fft_data = np.fft.fftshift(reshaped_data)
    # PCA   3  
    pca = PCA(n_components=n_components)
    pca_data = pca.fit_transform(fft_data)
    inverse_pca_data = pca.inverse_transform(pca_data)
    inverse_fft_data = np.fft.ifftshift(inverse_pca_data)
    filtered_data = np.array(inverse_fft_data).reshape(n_epochs, n_channels, n_times)
    filtered_epochs = mne.EpochsArray(filtered_data, epochs.info, tmin=epochs.tmin)
    return filtered_epochs

def spectr_filter(epochs):
    data = epochs.get_data()
    frequencies, times, Z = stft(data)
    power = np.abs(Z)**2
    threshold = np.percentile(power, 99.99)  #  25%    
    #   
    mask = power < threshold
    clean_Z = Z * mask
    #   
    _, clean_data = istft(clean_Z)
    clean_epochs = mne.EpochsArray(clean_data, epochs.info)
    return clean_epochs

def emd_decompose(epochs, evoked):
    EEG = epochs.copy().get_data() * 10**6
    ERP = evoked.copy().get_data() * 10**6
    n_epochs, n_channels, n_times = EEG.shape
    sf = int(epochs.info['sfreq'])

    hilbert_erp = []
    ch_names = epochs.ch_names

    for pick_idx, ch_name in enumerate(ch_names):
        filtered_electrode_data = []
        max_corr_data = []
        for ep in range(n_epochs):
            data = EEG[ep][pick_idx]
            erp_data = ERP[pick_idx]
            imfs = emd.sift.mask_sift(data, max_imfs=5)

            def find_optimal_imfs(imfs, erp_data):
                n_imfs = imfs.shape[1]
                max_corr = -1
                best_combination = None
                for r in range(1, n_imfs + 1):
                    for combination in combinations(range(n_imfs), r):
                        combined_imf = np.sum(imfs[:, combination], axis=1)
                        corr = np.corrcoef(erp_data, combined_imf)[0, 1]
                        if corr > max_corr:
                            max_corr = corr
                            best_combination = combination
                return best_combination, max_corr

            optimal_imfs, max_correlation = find_optimal_imfs(imfs, erp_data)
            filtered_signal = np.sum(imfs[:, optimal_imfs], axis=1)
            filtered_electrode_data.append(filtered_signal)
            max_corr_data.append(max_correlation)

        #       
        std_deviation = np.std(filtered_electrode_data, axis=0)
        standard_error = std_deviation / np.sqrt(n_epochs)
        mean_standard_error = np.mean(standard_error)
        print(f'Electrode {ch_names[pick_idx]}')
        print(f"\tMean Standard Error={mean_standard_error:.3f}")
        print(f'\tCorrelation: mean={np.mean(max_corr_data):.3f}, std={np.std(max_corr_data, ddof=1):.3f}')

        hilbert_erp.append(filtered_electrode_data)

    #       
    hilbert_erp = np.array(hilbert_erp)
    #      (n_epochs, n_channels, n_times)
    hilbert_erp = hilbert_erp.transpose(1, 0, 2) / 10**6
    #    EpochsArray   
    epochs_denoised = mne.EpochsArray(hilbert_erp, epochs.info, tmin=epochs.tmin, 
                                      baseline=epochs.baseline, tmax=epochs.tmax, verbose=False)
    return epochs_denoised



from setuptools import setup, find_packages
import pathlib

here = pathlib.Path(__file__).parent.resolve()
long_description = (here / "README.md").read_text(encoding="utf-8")
def parse_requirements(filename):
    with open(filename, 'r', encoding='utf-8') as f:
        return f.read().splitlines()


setup(
    name="eeg_auto_tools",
    version="0.0.7",
    author="Sear",
    license="Apache-2.0",
    author_email="vasilijkrukovskij2015@gmail.com",
    description="The set of tools for working with EEG data",
    packages=find_packages(),
    install_requires=parse_requirements('requirements.txt'),
    include_package_data=True,
    long_description=long_description,
    long_description_content_type='text/markdown',
    python_requires=">=3.6",
)


# Copyright 2025 Sear Gamemode
import numpy as np 
import mne 
import re 

def get_ANT_ROI():
    P1_ROI = ['P3', 'Pz', 'P4', 'POz', 'P1', 'P2']
    N1_ROI = ['P3', 'Pz', 'P4', 'O1', 'Oz', 'O2']
    CNV_ROI = ['FCz', 'Cz']
    N2_ROI = ['P3', 'Pz', 'P4', 'POz', 'P1', 'P2']
    P3_ROI = ['P3', 'Pz', 'P4', 'POz', 'P1', 'P2']
    COMPONENTS_ROI = {'P1': P1_ROI,
           'N1': N1_ROI,
           'CNV': CNV_ROI,
           'N2': N2_ROI,
           'P3': P3_ROI}
    return COMPONENTS_ROI

def make_ANT_events(raw, target_stimulus):
    events, event_id = mne.events_from_annotations(raw, verbose=False)
    desired_stimulus = 'Stimulus/s200'
    target_stimuli_codes = dict(list(zip(event_id.values(), event_id.keys())))
    filtered_events = []
    for i in range(0, len(events) - 1):
        current_event = events[i]
        next_event = events[i + 1]
        if target_stimuli_codes[current_event[2]] in target_stimulus:
            if target_stimuli_codes[next_event[2]] == desired_stimulus:
                filtered_events.append(current_event)
    filtered_event_id = {stimulus: event_id[stimulus] for stimulus in target_stimulus if stimulus in event_id}
    filtered_events = np.array(filtered_events)
    return filtered_event_id, filtered_events

def make_CB_events(raw, mode):
    events, event_id = mne.events_from_annotations(raw, verbose=False)
    event_id_rev = {v: k for k, v in event_id.items()}

    if mode == "Rest-IAT":
        reg = "cb"
        cb_events = [event for event in events if reg in event_id_rev[event[2]]]

    else:
        reg = "tar"
        cb_events = []
        seen_rn_values = set()

        for event in events:
            event_code = event_id_rev[event[2]]
            if (reg in event_code):
                rn_value = None
                match = re.search(r"rn(\d+)", event_code)
                if match:
                    rn_value = match.group(1)

                if rn_value and rn_value not in seen_rn_values:
                    cb_events.append(event)
                    seen_rn_values.add(rn_value)

    cb_events = np.array(cb_events)

    unique_cb_labels = list(set(event_id_rev[event[2]] for event in cb_events))
    cb_event_id = {label: event_id[label] for label in unique_cb_labels}
    return cb_events, cb_event_id

def make_RiTi_events(raw, stimulus_list, filt=False):
    #    
    events, event_id = mne.events_from_annotations(raw, verbose=False)
    inverted_dict = {v: k for k, v in event_id.items()}  #  event_id  
    if filt:
        filtered_events = []
        for event in events:
            stimulus = inverted_dict[event[2]]
            if stimulus in stimulus_list:
                filtered_events.append(event)
        #     numpy 
        filtered_events = np.array(filtered_events, dtype=int)
    else:    
        filtered_events = np.array(events, dtype=int)

    #     -       
    duplet_events = []
    duplet_names = {}
    
    duplet_code = 1  #    

    #   ,    (index 1)
    for i in range(1, len(filtered_events)):
        current_stimulus = inverted_dict[filtered_events[i, 2]]  #  
        previous_stimulus = inverted_dict[filtered_events[i - 1, 2]]  #  

        # ,          
        if (current_stimulus in stimulus_list) and (previous_stimulus in stimulus_list) and (current_stimulus != previous_stimulus):
            #    , , 'Stimulus/s11_s12'
            duplet_name = f'{previous_stimulus[:]}_{current_stimulus[:]}'
            #        ,   
            if duplet_name not in duplet_names:
                duplet_names[duplet_name] = duplet_code
                duplet_code += 1

            #       
            duplet_events.append([filtered_events[i, 0], 0, duplet_names[duplet_name]])

    #  duplet_events  numpy     MNE
    duplet_events = np.array(duplet_events, dtype=int)

    #  event_id  
    event_id_duplets = {name: code for name, code in duplet_names.items()}

    return duplet_events, event_id_duplets




# Copyright 2025 Sear Gamemode
import re 
from datetime import datetime
import os 
import mne 
import numpy as np 

electrodesofmontages = [
    'Fp1', 'Fpz', 'Fp2', 'F7', 'F3', 'Fz', 'F4', 'F8', 'FC5', 'FC1', 'FC2', 'FC6',
    'M1', 'T7', 'C3', 'Cz', 'C4', 'T8', 'M2', 'CP5', 'CP1', 'CP2', 'CP6', 'P7',
    'P3', 'Pz', 'P4', 'P8', 'POz', 'O1', 'Oz', 'O2', 'AF7', 'AF3', 'AF4', 'AF8',
    'F5', 'F1', 'F2', 'F6', 'FC3', 'FCz', 'FC4', 'C5', 'C1', 'C2', 'C6', 'CP3',
    'CP4', 'P5', 'P1', 'P2', 'P6', 'PO5', 'PO3', 'PO4', 'PO6', 'FT7', 'FT8',
    'TP7', 'TP8', 'PO7', 'PO8']

def get_riti_seq():
    sequence_file = 'RiTi_seq.txt'
    seq_list = []
    transform = {'Stimulus/s115':1, 'Stimulus/s120':4, 'Stimulus/s130':2, 'Stimulus/s160':3, 'Stimulus/s240':5}
    reverse_transform = dict(zip(transform.values(), transform.keys()))
    with open(sequence_file, 'r') as file:
        for line in file:
            seq_list.append(reverse_transform[int(line)])
    return seq_list

def get_stim(scenarious_name):
    list_stim = None 
    if scenarious_name == 'ANT':
        list_stim = ['Stimulus/s143', 'Stimulus/s145', 'Stimulus/s149', 'Stimulus/s151', 
                     'Stimulus/s141', 'Stimulus/s147', 'Stimulus/s153', 'Stimulus/s155']
    elif scenarious_name == 'RiTi':
        list_stim = ['Stimulus/s115', 'Stimulus/s120', 'Stimulus/s130', 'Stimulus/s160', 'Stimulus/s240']
    return list_stim

def get_meta(file_name):
    file_pattern = re.compile(
        r'INP(?P<id>\d{4})_v1\.(?P<visit_num>\d)_(?P<experiment>[^_]+)_'
        r'(?P<operator_code>[A-Za-z0-9_]+)_(?P<date>\d{2}\.\d{2}\.\d{2,4})'
        r'\.(?P<format>vhdr|vmrk|eeg)')
    match = file_pattern.match(file_name)
    return match.groupdict() if match else None

def get_brainvision_files(vhdr_path):
    with open(vhdr_path, 'r', encoding='utf-8') as f:
        lines = f.readlines()
    #  
    eeg_fname = None
    vmrk_fname = None
    #     
    for line in lines:
        line = line.strip()
        if line.startswith('DataFile='):
            eeg_fname = line.split('=')[1].strip()
        elif line.startswith('MarkerFile='):
            vmrk_fname = line.split('=')[1].strip()
    return (eeg_fname, vmrk_fname)


def get_file_info(file_path, elc_path):
    #   EEG
    raw = mne.io.read_raw_brainvision(file_path, 
                                      preload=False, verbose=False)
    #  -
    eeg_files = (file_path,) + get_brainvision_files(file_path)
    eeg_files = list(map(os.path.basename, eeg_files))
    meta = get_meta(eeg_files[0])
    #   EEG
    meas_date = raw.info['meas_date']
    if meas_date is not None:
        if isinstance(meas_date, datetime):
            measurement_datetime = meas_date.strftime('%Y-%m-%d %H:%M:%S')
    else:
        measurement_datetime = 'Unknown'
    #    
    data = {
        "id": meta.get('id', 'N/A'),
        "visit_num": meta.get('visit_num', 'N/A'),
        "scenario": meta.get('experiment', 'N/A'),
        "operator_code": ', '.join(meta.get('operator_code', 'N/A').split('_')),
        "eeg_files": ' '.join(eeg_files),
        "elc_file": os.path.basename(elc_path) if elc_path else 'Not found',
        "raw_type": type(raw).__name__,
        "duration": f"{raw.times[-1]:.2f}",
        "nchan": raw.info['nchan'],
        "electrodes_of_montages" : set(electrodesofmontages).issubset(set(raw.ch_names)),
        "ecg_channel": 'BIP1' in raw.ch_names,
        "eog_channel": 'EOG' in raw.ch_names,
        "sfreq": raw.info['sfreq'],
        "measurement_datetime": measurement_datetime,
        "highpass": raw.info['highpass'],
        "lowpass": raw.info['lowpass'],
        "custom_ref_applied": raw.info.get('custom_ref_applied', 'Not available'),
    }
    return data


def is_subsequence(subseq, seq):
    subseq_index = 0
    subseq_len = len(subseq)
    flag = 0
    transitions = 0
    for elem in seq:
        copy_flag = flag
        if subseq_index < subseq_len and elem == subseq[subseq_index]:
            subseq_index += 1
            flag = 1
        else:
            flag = 0
        if flag != copy_flag:
            transitions += 1
    return subseq_index == subseq_len, subseq_index, transitions

def verificate_events(raw, seq_list, shift=1.85, var=0.35):
    events, event_id = mne.events_from_annotations(raw, verbose=False)
    reverse_event_id = dict(zip(event_id.values(), event_id.keys()))
    
    tar_idx = 0
    new_events = []
    
    n_events = len(events)
    n_seq = len(seq_list)
    events_times = events[:, 0]/raw.info['sfreq']
    events_seq = np.array([reverse_event_id[events[i, 2]] for i in range(len(events))])
    current_time = events_times[0] - shift

    #new_events.append(events[0])
    while not (tar_idx > n_seq - 1 or current_time > events_times[-1]):
        
        next_time_l = current_time + shift - var
        next_time_r = current_time + shift + var

        mask = (events_times >= next_time_l) & (events_times <= next_time_r)
        filtered_seq = events_seq[mask]
        filtered_events = events[mask]
        filtered_times = events_times[mask]
        if filtered_seq.size > 0:
            unique_elements = np.unique(filtered_seq)
            if len(unique_elements) == 1:
                new_events.append(filtered_events[0])
                tar_idx += 1
                current_time = filtered_times[0]
            else:
                matches = seq_list[tar_idx] == filtered_seq
                if matches.all():
                    new_events.append(filtered_events[matches[0]])
                    tar_idx += 1
                    current_time = filtered_times[matches[0]]
                else:    
                    tar_idx += 1
                    current_time += shift
        else:
            tar_idx += 1
            current_time += var
    
    return np.array(new_events)
